MODULE Compiler;

(* --------------------------------------------------------------------------
 * (C) 2008 by Alexander Iljin
 * -------------------------------------------------------------------------- *)

IMPORT SYSTEM,Win:=Windows,Convert,S:=Str;

(** -------------------------------------------------------------------------
  * This module implements command-line compiler integration.
  * ------------------------------------------------------------------------- *)

TYPE
   Str* = ARRAY 81 OF CHAR; (* console width + 1 *)
   Error* = POINTER TO RECORD
      file-: Str;     (** File name the error reported for (no path)  *)
      msg- : Str;     (** Error message reported by the compiler      *)
      line-: INTEGER; (** Line containing the error (starting with 1) *)
      col- : INTEGER; (** Error position in the line (leftmost = 0)   *)
   END;
   (** The errors generated by this module are reported as Error records with
     * line = -1, col = -1 and file = ''. *)

   Pipe = RECORD
      rd: Win.HANDLE;
      wr: Win.HANDLE;
   END;

   Process = RECORD
      proc: Win.HANDLE;
      thd: Win.HANDLE;
   END;

PROCEDURE ClearMem (VAR mem: ARRAY OF SYSTEM.BYTE);
VAR i: LONGINT;
BEGIN
   i := LEN (mem);
   WHILE i > 0 DO
      DEC (i);
      mem [i] := 0;
   END;
END ClearMem;

PROCEDURE CreatePipe (VAR p: Pipe): BOOLEAN;
(* Create a pipe with inheritable handles. If buffSize = 0, let the system
 * decide.  *)
CONST buffSize = 4 * SIZE (Str);
VAR
   sec: Win.SECURITY_ATTRIBUTES;
   ptr: Win.PSECURITY_ATTRIBUTES;
BEGIN
   sec.nLength := SIZE (sec);
   sec.lpSecurityDescriptor := NIL;
   sec.bInheritHandle := Win.TRUE;
   ptr := SYSTEM.VAL (Win.PSECURITY_ATTRIBUTES, SYSTEM.ADR (sec));
   RETURN Win.CreatePipe (p.rd, p.wr, ptr, buffSize);
END CreatePipe;

PROCEDURE CreateWritePipe (VAR p: Pipe): BOOLEAN;
BEGIN
   RETURN CreatePipe (p)
      & Win.SetHandleInformation (p.rd, Win.HANDLE_FLAG_INHERIT, Win.HANDLE_FLAG_SET {});
END CreateWritePipe;

PROCEDURE CreateReadPipe (VAR p: Pipe): BOOLEAN;
BEGIN
   RETURN CreatePipe (p)
      & Win.SetHandleInformation (p.wr, Win.HANDLE_FLAG_INHERIT, Win.HANDLE_FLAG_SET {});
END CreateReadPipe;

PROCEDURE ClosePipe (VAR pipe: Pipe);
BEGIN
   Win.CloseHandle (pipe.rd);
   Win.CloseHandle (pipe.wr);
END ClosePipe;

PROCEDURE CreateProcess (name, currPath: ARRAY OF CHAR; VAR wrPipe: Pipe; VAR proc: Process): BOOLEAN;
VAR
   pname,ppath: Win.PCSTR;
   si: Win.STARTUPINFO;
   pi: Win.PROCESS_INFORMATION;
   rdPipe: Pipe;
   res: BOOLEAN;
BEGIN
   res := CreateReadPipe (rdPipe);
   IF res THEN
      pname := SYSTEM.VAL (Win.PCSTR, SYSTEM.ADR (name));
      ppath := SYSTEM.VAL (Win.PCSTR, SYSTEM.ADR (currPath));
      ClearMem (si);
      si.cb := SIZE (si);
      si.dwFlags := Win.STARTF_USESTDHANDLES + Win.STARTF_USESHOWWINDOW;
      si.hStdInput := rdPipe.rd;
      si.hStdOutput := wrPipe.wr;
      si.hStdError := wrPipe.wr;
      si.wShowWindow := SYSTEM.VAL (Win.WORD, Win.SW_SHOWMINNOACTIVE);
      res := Win.CreateProcess (NIL, pname, NIL, NIL, TRUE, Win.CREATE_SET {}, NIL, ppath, si, pi);
      IF res THEN
         proc.proc := pi.hProcess;
         proc.thd := pi.hThread;
         (* the wrPipe write handle is transferred, we should close our copy *)
         Win.CloseHandle (wrPipe.wr);
         wrPipe.wr := NIL;
      END;
      (* the rdPipe read handle is transferred, we should close our copy
       * plus, we're not going to write there anything *)
      ClosePipe (rdPipe);
   END;
   RETURN res
END CreateProcess;

PROCEDURE CloseProcess (VAR proc: Process);
BEGIN
   Win.CloseHandle (proc.proc);
   Win.CloseHandle (proc.thd);
END CloseProcess;

PROCEDURE ExtractPath (VAR fname: ARRAY OF CHAR);
CONST delim = '\';
VAR i,lastDelim: LONGINT;
BEGIN
   i := 0; lastDelim := -1;
   WHILE fname [i] # 0X DO
      IF fname [i] = delim THEN
         lastDelim := i;
      END;
      INC (i);
   END;
   IF lastDelim > 0 THEN
      fname [lastDelim + 1] := 0X;
   END;
END ExtractPath;

PROCEDURE DumpPipe (VAR pipe: Pipe; VAR path: ARRAY OF CHAR): BOOLEAN;
(* Read all data from the pipe and output it into the file. Return FALSE if
 * the output file could not be created. *)
CONST fname = 'pipedump.txt';
VAR
   fn  : ARRAY Win.MAX_PATH OF CHAR;
   file: Win.HANDLE;
   buff: ARRAY 1024 OF SYSTEM.BYTE;
   read: Win.DWORD;
   res : BOOLEAN;
BEGIN
   COPY (path, fn);
   S.AppendC (fn, fname);
   file := Win.CreateFile (fn, Win.GENERIC_WRITE, Win.FILE_SHARE_READ,
      NIL, Win.CREATE_ALWAYS, Win.FILE_ATTRIBUTE_NORMAL, NIL);
   res := file # Win.INVALID_HANDLE_VALUE;
   IF res THEN
      WHILE Win.ReadFile (pipe.rd, buff, LEN (buff), read, NIL)
         & Win.WriteFile (file   , buff, read      , read, NIL)
      DO
      END;
      Win.CloseHandle (file);
   END;
   RETURN res
END DumpPipe;

PROCEDURE DrainPipe (VAR pipe: Pipe);
(* Read and discard all data from the pipe. *)
VAR
   buff: ARRAY 1024 OF SYSTEM.BYTE;
   read: Win.DWORD;
BEGIN
   WHILE Win.ReadFile (pipe.rd, buff, LEN (buff), read, NIL) DO
   END;
END DrainPipe;

PROCEDURE ReadLn (VAR pipe: Pipe; VAR line: ARRAY OF CHAR; VAR more: BOOLEAN);
(* Read one line of text from the pipe, imitating console: LF terminates the
 * line, CR only returns the caret to the beginning of a line. Return more =
 * TRUE if there is more data left to be read for subsequent calls. *)
VAR
   i : INTEGER;
   ch: ARRAY 1 OF CHAR;
   read: Win.DWORD;
BEGIN
   ClearMem (line);
   i := 0; more := TRUE; ch [0] := 0X;
   WHILE (i < LEN (line) - 1) & more & (ch [0] # 0AX) DO
      more := Win.ReadFile (pipe.rd, ch, 1, read, NIL) & (read = 1);
      IF more & (ch [0] # 0AX) THEN
         IF ch [0] = 0DX THEN
            i := 0;
         ELSE
            line [i] := ch [0];
            INC (i);
         END;
      END;
   END;
END ReadLn;

PROCEDURE ParseCompilerOutput (VAR pipe: Pipe; VAR err: Error);
VAR
   again, getMsg: BOOLEAN;
   line: Str;

   PROCEDURE GetErrorCoords (): BOOLEAN;
   VAR
      i,c: LONGINT;
      num: Convert.NumberStr;
   BEGIN
      c := S.CharPosFrom (' ', line, 3);
      IF c = S.NotFound THEN RETURN FALSE END;
      S.CopyPart (line, err.file, 3, c - 1);
      i := c + 1;
      c := S.CharPosFrom ('.', line, i);
      IF c = S.NotFound THEN RETURN FALSE END;
      S.CopyPart (line, num, i, c - 1);
      IF ~Convert.StrToInt (num, err.line) THEN RETURN FALSE END;
      i := c + 1;
      c := S.CharPosFrom (' ', line, i);
      IF c = S.NotFound THEN RETURN FALSE END;
      S.CopyPart (line, num, i, c - 1);
      IF ~Convert.StrToInt (num, err.col) THEN RETURN FALSE END;
      RETURN TRUE
   END GetErrorCoords;

BEGIN
   getMsg := FALSE;
   REPEAT
      ReadLn (pipe, line, again);
      IF (line [0] = '*') & (line [1] = ' ') THEN
         IF ~getMsg THEN
            IF line [2] = '[' THEN
               getMsg := GetErrorCoords ();
            END
         ELSE
            getMsg := FALSE;
            S.CopyPart (line, err.msg, 2, LEN (err.msg));
            again := FALSE;
         END;
      END;
   UNTIL ~again;
   (* read the rest of the data, otherwise XDS throws exception *)
   DrainPipe (pipe);
END ParseCompilerOutput;

PROCEDURE Compile* (VAR fname: ARRAY OF CHAR): Error;
(** Try to compile the project determined by the fname, return an Error or NIL
  * on success. *)
CONST
   CompileCmd = 'make';
VAR
   pipe: Pipe;
   proc: Process;
   res: Error;
BEGIN
   NEW (res); res.file := ''; res.msg := ''; res.line := -1; res.col := -1;
   ExtractPath (fname);
   IF CreateWritePipe (pipe) THEN
      IF CreateProcess (CompileCmd, fname, pipe, proc) THEN
         ParseCompilerOutput (pipe, res);
         -- IF ~DumpPipe (pipe, fname) THEN
            -- res.msg := "Can't create dump file.";
         -- END;
         CloseProcess (proc);
      ELSE
         res.msg := "Can't create compiler process";
      END;
      ClosePipe (pipe)
   ELSE
      res.msg := "Can't create communication pipes";
   END;
   IF res.msg = '' THEN
      res := NIL;
   END;
   RETURN res
END Compile;

END Compiler.
