MODULE Settings;

(* --------------------------------------------------------------------------
 * (C) 2010 by Alexander Iljin
 * -------------------------------------------------------------------------- *)

IMPORT Convert, Files, Str, Sci:=Scintilla, Streams;

(** -------------------------------------------------------------------------
  * This module keeps various global settings of the application.
  * ------------------------------------------------------------------------- *)

CONST
   GlobalSection*   = 'Global';
   LookupKey*       = 'lookup';
   AutoCHeightKey   = 'autoCMaxHeight';
   MouseHintDelayKey= 'mouseHintDelay';
   LookupSeparator* = ';';
   AutoCMaxHeight   = 32; (* Default value for autoCMaxHeight *)
   MouseHintDelay   = 500; (* Default value for mouseHintDelay *)

VAR
   lookup-: Str.PTable; (* module lookup paths *)
   autoCMaxHeight-: LONGINT; (* The maximum number of lines in the autocompletion box *)
   mouseHintDelay-: LONGINT; (* Code Hint under mouse cursor delay in msec *)

PROCEDURE SetLookupTable* (VAR str: ARRAY OF CHAR);
(** Assign the list of paths to the 'lookup' table. 'str' format:
   [path1[;path2...]] *)
BEGIN
   lookup := NIL;
   Str.DynLoadTable (str, LookupSeparator, lookup);
END SetLookupTable;

PROCEDURE LoadFromStream* (stream: Streams.Stream);
CONST
   CommentChar = ';';
VAR
   line, key, value: ARRAY 1024 OF CHAR;
   equalPos: LONGINT;
   eof, section: BOOLEAN;

   PROCEDURE ReadLine (): BOOLEAN;
   (* Read until a non-empty line is read, return TRUE on success. If a new
    * section header is found, return FALSE, line will contain the section
    * string "[section name]", section = TRUE, otherwise section = FALSE. *)
   CONST MinUsefulLen = 2; (* shorter line can't be a value or a section *)
   VAR len: LONGINT; (* useful 'line' contents length, = 0 for comments *)
   BEGIN
      section := FALSE;
      REPEAT
         stream.ReadLn (line);
         IF stream.err # Streams.OK THEN
            len := 0;
            eof := TRUE;
         ELSIF line [0] = CommentChar THEN (* comment *)
            len := 0;
         ELSE
            len := Str.Length (line);
            IF len < MinUsefulLen THEN
               len := 0;
            ELSE
               section := (line [0] = '[') & (line [len - 1] = ']');
               IF ~section THEN (* get key and value *)
                  equalPos := 0;
                  WHILE (equalPos < len) & ~(line [equalPos] = '=') DO
                     INC (equalPos);
                  END;
                  IF equalPos < len THEN (* found *)
                     Str.CopyPart (line, key, 0, equalPos - 1);
                     Str.CopyPart (line, value, equalPos + 1, len);
                  ELSE
                     len := 0;
                  END;
               END;
            END;
         END;
      UNTIL eof OR section OR (len > 0);
      line [len] := 0X;
      RETURN (len > 0) & ~section
   END ReadLine;

   PROCEDURE ValueToLongDef (VAR res: LONGINT; min, max, default: LONGINT);
   VAR
      i: LONGINT;
   BEGIN
      IF Convert.StrToLong (value, i) & (i >= min) & (i <= max) THEN
         res := i;
      ELSE
         res := default;
      END;
   END ValueToLongDef;

BEGIN (* LoadFromStream *)
   ASSERT (stream # NIL, 20);
   eof := FALSE;
   WHILE ReadLine () DO
   END;
   WHILE section DO
      IF line = '[' + GlobalSection + ']' THEN
         WHILE ReadLine () DO
            IF key = LookupKey THEN
               SetLookupTable (value);
            ELSIF key = AutoCHeightKey THEN
               ValueToLongDef (autoCMaxHeight, 1, MAX (LONGINT), AutoCMaxHeight);
            ELSIF key = MouseHintDelayKey THEN
               ValueToLongDef (mouseHintDelay, 0, Sci.timeForever, MouseHintDelay);
            END;
         END;
      ELSE (* skip unknown section *)
         WHILE ReadLine () DO
         END;
      END;
   END;
END LoadFromStream;

PROCEDURE LoadFromFile* (VAR name: ARRAY OF CHAR);
VAR file: Streams.Stream;
BEGIN
   IF Files.OpenBuffered (file, name, Files.ReadMode + Files.DenyWrite, FALSE, FALSE) THEN
      LoadFromStream (file);
      file.Close;
   END;
END LoadFromFile;

PROCEDURE Init;
CONST DefaultLookupPaths = '.\;..\A3;..\Lib;..\..\A3;..\..\Lib;..\..\Src';
VAR tmp: ARRAY LEN (DefaultLookupPaths) OF CHAR;
BEGIN
   tmp := DefaultLookupPaths;
   SetLookupTable (tmp);
   autoCMaxHeight := AutoCMaxHeight;
   mouseHintDelay := MouseHintDelay;
END Init;

BEGIN
   Init;
END Settings.
