MODULE Tools;

(* --------------------------------------------------------------------------
 * (C) 2008 by Alexander Iljin
 * -------------------------------------------------------------------------- *)

IMPORT Project,Streams,Sci:=Scintilla,Str,Win:=Windows,Paths,Files,OPM,OPS,
   Npp, Settings;

(** -------------------------------------------------------------------------
  * Tool module contains various utility procedures and hides implementation
  * details of Scintilla handling, etc.
  * ------------------------------------------------------------------------- *)

TYPE
   BufferStr = ARRAY 256 OF CHAR;
   Path* = ARRAY Win.MAX_PATH OF CHAR;

   Ident* = POINTER TO IdentDesc;
   IdentDesc* = RECORD
      name-: Project.Name; (** identifier name                               *)
      next-: Ident;        (** next identifier in qualified identifier       *)
      pos-: LONGINT;       (** first letter position within the document     *)
   END;

   (* Reader implementation reading from Sterams.Stream *)
   StreamReader = POINTER TO RECORD (Project.ReaderDesc)
      s: Streams.Stream;
   END;

   (* Reader implementation reading from Scintilla component *)
   ScintillaReader = POINTER TO RECORD (Project.ReaderDesc)
      sci: Sci.Handle;  (* Scintilla component to read from              *)
      buff: BufferStr;  (* buffer of data read from Scintilla            *)
      buffPos: LONGINT; (* current reading position in buff              *)
      pos: LONGINT;     (* position of buff within Scintilla document    *)
      ch: CHAR;         (* last character read by ReadPrev               *)
   END;

   (* Filesystem-based Locator implementation. *)
   Locator* = POINTER TO LocatorDesc;
   LocatorDesc = RECORD (Project.LocatorDesc)
      path-: Path;      (* full path to the module file                  *)
      time: Win.FILETIME; (* file last write time                        *)
   END;

PROCEDURE (i: Ident) Length* (): LONGINT;
(** Return the length of the i.name field upto the 0X terminator. *)
BEGIN RETURN Str.Length (i.name)
END Length;

PROCEDURE (r: StreamReader) Read2 (VAR ch: CHAR);
(* Read the next character from the stream. *)
BEGIN
   r.s.ReadByte (ch);
   r.eot := r.s.err # Streams.OK;
   IF r.eot THEN
      r.s.Close;
   END;
END Read2;

PROCEDURE (r: ScintillaReader) Init ();
BEGIN
   r.Init^;
   r.pos := 0;
   r.buff := '';
   r.buffPos := -1;
   r.ch := 0X;
END Init;

PROCEDURE (r: ScintillaReader) Read2 (VAR ch: CHAR);
(* Read the next character from the Scintilla document. The read string is
 * buffered in r.buff. *)
VAR to, max: LONGINT;
BEGIN
   IF r.eot THEN
      ch := 0X
   ELSE
      INC (r.buffPos);
      IF r.buff [r.buffPos] = 0X THEN
         IF (r.buffPos = LEN (r.buff) - 1) OR (r.buffPos = 0) THEN
            INC (r.pos, r.buffPos);
            max := Sci.GetLength (r.sci);
            to := r.pos + LEN (r.buff) - 1;
            IF to > max THEN to := max END;
            r.buffPos := Sci.GetTextRange (r.sci, r.pos, to, r.buff);
            IF r.buffPos = 0 THEN
               r.buff := '';
               r.eot := TRUE;
            ELSE
               r.buff [r.buffPos] := 0X;
               r.buffPos := 0;
            END
         ELSE
            r.eot := TRUE
         END
      ELSE
      END;
      ch := r.buff [r.buffPos];
   END;
END Read2;

PROCEDURE (r: ScintillaReader) ReadFrom (pos: LONGINT);
(* Setup the reader so that the next Read call will read from position pos. *)
BEGIN
   ASSERT (pos >= 0, 20);
   r.eot := pos >= Sci.GetLength (r.sci);
   IF ~r.eot THEN
      IF pos = 0 THEN
         r.pos := 0;
         r.buffPos := -1;
      ELSE
         r.buffPos := LEN (r.buff) - 2;
         r.pos := pos - (LEN (r.buff) - 2) - 1;
      END;
      r.buff := '';
   END;
END ReadFrom;

PROCEDURE (r: ScintillaReader) ReadPrev ();
(* Read the previous character from the Scintilla document. The read string is
 * buffered in r.buff. *)
VAR to: LONGINT;
BEGIN
   IF r.eot THEN
      r.ch := 0X
   ELSE
      IF r.buffPos = 0 THEN
         IF r.pos > 0 THEN
            IF r.pos > LEN (r.buff) - 1 THEN
               DEC (r.pos, LEN (r.buff) - 1);
               to := r.pos + LEN (r.buff) - 1;
            ELSE
               to := r.pos;
               r.pos := 0;
            END;
            r.buffPos := Sci.GetTextRange (r.sci, r.pos, to, r.buff);
         ELSE
            r.eot := TRUE;
            r.ch := 0X;
         END
      END;
      IF r.buffPos > 0 THEN
         DEC (r.buffPos);
         r.ch := r.buff [r.buffPos];
      ELSE
         r.eot := TRUE;
         r.ch := 0X;
      END;
   END;
END ReadPrev;

PROCEDURE ParseStream* (s: Streams.Stream; VAR m: Project.Module);
(** Call Parse using the s stream as the data source. *)
VAR r: StreamReader;
BEGIN
   NEW (r); r.Init; r.s := s;
   Project.Parse (r, m);
END ParseStream;

PROCEDURE ParseScintilla* (sci: Sci.Handle; VAR m: Project.Module);
(** Call Parse using the sci Scintilla component as the data source. *)
VAR r: ScintillaReader;
BEGIN
   NEW (r); r.Init; r.sci := sci;
   Project.Parse (r, m);
END ParseScintilla;

PROCEDURE GetFileWriteTime (VAR name: ARRAY OF CHAR; VAR time: Win.FILETIME);
(* Return the file's last write time, or zero on error. *)
VAR
   file: Win.HANDLE;
   dummy: Win.FILETIME;
   ok: BOOLEAN;
BEGIN
   file := Win.CreateFile (name, Win.GENERIC_READ,
      Win.FILE_SHARE_READ + Win.FILE_SHARE_WRITE + Win.FILE_SHARE_DELETE,
      NIL, Win.OPEN_EXISTING, Win.FILE_ATTRIBUTE_NORMAL, NIL
   );
   ok := file # NIL;
   IF ok THEN
      ok := Win.GetFileTime (file, dummy, dummy, time);
      Win.CloseHandle (file);
   END;
   IF ~ok THEN
      time.dwLowDateTime := 0;
      time.dwHighDateTime := 0;
   END;
END GetFileWriteTime;

PROCEDURE (loc: Locator) MarkAsUpdated ();
(** Remember the current state of the file as being the most recent.
  * If the file was not modified after this procedure is called,
  * loc.RequiresUpdate should return FALSE. *)
BEGIN
   GetFileWriteTime (loc.path, loc.time);
END MarkAsUpdated;

PROCEDURE (loc: Locator) RequiresUpdate (): BOOLEAN;
(** Return TRUE if the file was modified since the last call to
  * loc.MarkAsUpdated, or if the file does not exist anymore. *)
VAR time: Win.FILETIME;
BEGIN
   IF (loc.time.dwLowDateTime = 0) & (loc.time.dwHighDateTime = 0) THEN
      RETURN TRUE
   ELSE
      GetFileWriteTime (loc.path, time);
      RETURN (loc.time.dwLowDateTime # time.dwLowDateTime)
         OR (loc.time.dwHighDateTime # time.dwHighDateTime)
   END;
END RequiresUpdate;

PROCEDURE (loc: Locator) SameAs (loc2: Project.Locator): BOOLEAN;
(** Return TRUE if loc and loc2 point to the same file. *)
VAR res: BOOLEAN;
BEGIN
   WITH loc2: Locator DO
      res := (loc = loc2) OR Paths.CaseSensitive & (loc.path = loc2.path)
         OR ~Paths.CaseSensitive & Str.UpperEqual (loc.path, loc2.path)
   ELSE
      res := FALSE;
   END;
   RETURN res
END SameAs;

PROCEDURE (loc: Locator) NewReader (): Project.Reader;
(* Create and return the file reader. *)
VAR
   res: StreamReader;
   s: Streams.Stream;
BEGIN
   IF Files.OpenBuffered (s, loc.path, Files.ReadMode, FALSE, FALSE) THEN
      NEW (res); res.Init; res.s := s;
   ELSE
      res := NIL;
   END;
   RETURN res
END NewReader;

PROCEDURE IdentChar* (ch: CHAR): BOOLEAN;
(* Return TRUE if ch is a valid character of an identifier *)
BEGIN
   RETURN ('a' <= ch) & (ch <= 'z')
      OR ('A' <= ch) & (ch <= 'Z')
      OR ('0' <= ch) & (ch <= '9')
      OR (ch = '_')
END IdentChar;

PROCEDURE IdentFirstChar* (ch: CHAR): BOOLEAN;
(* Return TRUE if ch is a valid first character of an identifier *)
BEGIN
   RETURN ('a' <= ch) & (ch <= 'z')
      OR ('A' <= ch) & (ch <= 'Z')
      OR (ch = '_')
END IdentFirstChar;

PROCEDURE MoveBlock (VAR str: ARRAY OF CHAR; from, len, to: LONGINT);
(* Move a block of "len" characters within "str" from position "from" to
 * position "to". *)
BEGIN
   WHILE len > 0 DO
      str [to] := str [from];
      INC (from); INC (to); DEC (len);
   END;
END MoveBlock;

PROCEDURE GetIdent (r: ScintillaReader; VAR id: IdentDesc);
(* Read the identifier name from the current position in the Reader. If the
 * read identifier is too long or starts with a digit, then set id.name =
 * empty string. *)
VAR i: LONGINT;
BEGIN
   i := LEN (id.name);
   WHILE (i > 0) & IdentChar (r.ch) DO
      DEC (i);
      id.name [i] := r.ch;
      r.ReadPrev;
   END;
   IF (i > 0) & (i < LEN (id.name)) & IdentFirstChar (id.name [i]) THEN
      MoveBlock (id.name, i, LEN (id.name) - i, 0);
      id.name [LEN (id.name) - i] := 0X
   ELSE
      id.name [0] := 0X
   END
END GetIdent;

PROCEDURE SkipBack (r: ScintillaReader);
(* Start from the current character and skip all whitespace and array index
 * specifications to the left. *)
VAR done: BOOLEAN;

   PROCEDURE SkipArrayIndex ();
   BEGIN
      REPEAT
         r.ReadPrev;
         IF r.ch = ']' THEN
            SkipArrayIndex
         END;
      UNTIL (r.ch = '[') OR r.eot;
   END SkipArrayIndex;

BEGIN
   done := FALSE;
   REPEAT
      IF ((r.ch <= ' ') OR (r.ch = '^')) & ~r.eot THEN
         r.ReadPrev
      ELSIF r.ch = ']' THEN
         SkipArrayIndex;
         r.ReadPrev;
      ELSE
         done := TRUE
      END
   UNTIL done;
END SkipBack;

PROCEDURE GetQualident (r: ScintillaReader): Ident;
VAR
   res, curr: Ident;
   done, space: BOOLEAN;
BEGIN
   res := NIL; curr := NIL;
   r.ReadPrev;
   space := r.ch = ' ';
   SkipBack (r);
   IF r.ch = '.' THEN
      NEW (res);
      res.name := '';
      res.pos := r.pos + r.buffPos + 1;
      r.ReadPrev;
      SkipBack (r);
   ELSIF space THEN
      RETURN res;
   END;
   IF IdentChar (r.ch) THEN
      done := FALSE;
      REPEAT
         IF curr = NIL THEN NEW (curr) END;
         GetIdent (r, curr^);
         IF curr.name [0] # 0X THEN
            curr.pos := r.pos + r.buffPos + 1;
            curr.next := res;
            res := curr;
            curr := NIL;
            SkipBack (r);
            IF r.ch = '.' THEN
               r.ReadPrev;
               SkipBack (r);
               done := ~IdentChar (r.ch);
            ELSE
               done := TRUE
            END
         ELSE
            done := TRUE
         END
      UNTIL done
   END;
   RETURN res;
END GetQualident;

PROCEDURE GetQualidentAt* (sc: Sci.Handle; pos: LONGINT): Ident;
(** Get a chain of identifiers separated by dots, e.g. "Module.variable", to
  * the left of a given character position in the Scintilla document. The
  * last identifier in the chain is returned upto the [pos] character.
  * Array index specifications are skipped. Type guards and procedure calls
  * are not supported (except when part of an array index).  *)
VAR r: ScintillaReader;
BEGIN
   ASSERT (pos >= 0, 20);
   NEW (r); r.Init (); r.sci := sc;
   r.buffPos := 0;
   r.pos := pos + 1;
   RETURN GetQualident (r)
END GetQualidentAt;

PROCEDURE ReadIdentFrom* (sc: Sci.Handle; pos: LONGINT): Ident;
(** Read a single identifier from position pos in document sc using the OPS
  * scanner. If there is no identifier at that position, return NIL. *)
VAR
   r  : ScintillaReader;
   sym: SHORTINT;
   res: Ident;
   ch : CHAR;
BEGIN
   ASSERT (pos >= 0, 20);
   res := NIL;
   NEW (r); r.Init (); r.sci := sc;
   r.ReadFrom (pos);
   r.Read2 (ch);
   IF ~r.eot & IdentFirstChar (ch) THEN
      r.buffPos := -1;
      OPM.Init (r); OPS.Init;
      OPS.Get (sym);
      IF sym = OPS.ident THEN
         NEW (res);
         res.name := OPS.name;
      END;
   END;
   RETURN res;
END ReadIdentFrom;

PROCEDURE GetModuleLocator* (VAR name: Project.Name; VAR currPath: Path): Locator;
(** Find the module "name" given that the current path is currPath. Create and
  * return a locator instance for the module. If the module file could not be
  * found, return NIL. *)
CONST moduleFileExt = 'ob2';
VAR
   fs  : Paths.Specifier;
   dir : Path; (* base directory, a copy of currPath *)
   path: Path; (* last checked full file name *)
   res : Locator;
   i,c : LONGINT;

   PROCEDURE LookupFile (tailDir: ARRAY OF CHAR): BOOLEAN;
   BEGIN
      COPY (dir, fs.dir);
      Str.Append (fs.dir, tailDir);
      Paths.RemoveRelative (fs);
      IF Paths.Exist (fs) THEN
         NEW (res);
         Paths.FullName (fs, path);
         COPY (path, res.path);
      END;
      RETURN res # NIL
   END LookupFile;

BEGIN
   res := NIL; Paths.Init (fs);
   IF Paths.Parse (currPath, fs, TRUE) THEN
      Paths.Complete (fs);
      COPY (name, fs.name);
      fs.ext := moduleFileExt;
      COPY (fs.dir, dir);
      i := 0;
      c := LEN (Settings.lookup^);
      WHILE (i < c) & ~LookupFile (Settings.lookup [i]^) DO
         INC (i);
      END
   END;
   RETURN res
END GetModuleLocator;

PROCEDURE OpenFile* (VAR fname: Path): BOOLEAN;
(** Open the requested file in Notepad++, return TRUE on success. *)
VAR
   curr: Path;
   res: BOOLEAN;
BEGIN
   Npp.GetFullCurrentPath (curr);
   res := Paths.CaseSensitive & (curr = fname)
      OR ~Paths.CaseSensitive & Str.UpperEqual (curr, fname)
      OR Npp.OpenFile (fname); (* returns FALSE if file is already open *)
   IF ~res THEN
      (* if Npp.OpenFile returned FALSE, it could have activated the
       * already-open file that we need *)
      Npp.GetFullCurrentPath (curr);
      res := Paths.CaseSensitive & (curr = fname)
         OR ~Paths.CaseSensitive & Str.UpperEqual (curr, fname)
   END;
   RETURN res
END OpenFile;

PROCEDURE OpenLoc* (loc: Locator): BOOLEAN;
(** Open the file pointed to by the given Locator. *)
BEGIN
   RETURN (loc # NIL) & OpenFile (loc.path)
END OpenLoc;

PROCEDURE OpenModule* (m: Project.Module): BOOLEAN;
(** Open the requested module by its Locator, return TRUE on success. *)
BEGIN
   RETURN (m # NIL) & (m.loc IS Locator) & OpenFile (m.loc (Locator).path)
END OpenModule;

PROCEDURE LengthC* (str: ARRAY OF CHAR): LONGINT;
(** Return the length of the string passed as a constant parameter. *)
BEGIN RETURN Str.Length (str);
END LengthC;

END Tools.
